package Block;

use warnings;

# use Nark; Nark::nark();

sub new { my $type = shift; bless [@_], $type; }

sub x1        (@) :lvalue { $_[0]->[0] }
sub x2        (@) :lvalue { $_[0]->[1] }
sub y1        (@) :lvalue { $_[0]->[2] }
sub y2        (@) :lvalue { $_[0]->[3] }
sub color     (@) :lvalue { $_[0]->[4] }
sub neighbors (@) :lvalue { $_[0]->[5] ||= []; $_[0]->[5] }
sub builder   (@) :lvalue { $_[0]->[6] }

sub add_neighbor { $_[1]->isa('Block') or die; push @{&neighbors}, $_[1] }
sub query_neighbors { wantarray ? @{&neighbors} : &neighbors }
sub query_x_dim { return &x2 - &x1; }
sub query_y_dim { return &y2 - &y1; }

sub get_color { 
  my $lets = ['A' .. 'Z', 'a' .. 'z', '1' .. '9']; 
  return $lets->[&color % @$lets]; 
}

sub solve_maze {
  my $me = $_[0];
  my $dest = $_[1];
  return ($me) if $me eq $dest;
  my @queue = ($me);
  my %route; $route{$me} = [$me];
  my %routeblock; $routeblocks{$me} = $me;
  while(my $ob = shift @queue) {
    # print "debug: at: ", $ob->get_color(), "  dest: ", $dest->get_color(), "\n";
    foreach my $i (@{$ob->neighbors()}) {
      # print "  exa: ", $i->get_color(), "\n";
      next unless $i;
      if(!$route{$i}) {
        $route{$i} = [ @{$route{$ob}}, $i ];
        $routeblocks{$i} = $i;
        return @{$route{$i}} if($i eq $dest);
        push @queue, $i;
      }
    }
  }
  warn "solve_maze couldn't solve maze: source is " . $me->get_color() . " target is " . 
       $dest->get_color() . "\n";
  print "blocks we know about:\n";
  foreach my $i (keys %route) {
    print $routeblocks{$i}->get_color(), ' @ ', $routeblocks{$i}->x1(), ',', 
          $routeblocks{$i}->y1(),
          '  : ', scalar(@{$route{$i}}), ' blocks to dest ', "\n";
  }
  return;
}

sub approx_cost {
  my $me = $_[0];
  my $dest = $_[1];
  # expected cost to traverse this route. penalizes "tunnels"
  my @route = $me->solve_maze($dest);
  return unless @route;
  my $sum; 
  foreach my $i (@route) {
     $sum += ($i->query_x_dim() + $i->query_y_dim()) / 2;
     $sum += $i->query_x_dim() * 5 if $i->query_y_dim() == 1;
     $sum += $i->query_y_dim() * 5 if $i->query_x_dim() == 1;
  }
  return $sum;
}

sub common_border {
  my $me = $_[0];
  my $dest = $_[1];
  my $least = sub { $_[0] < $_[1] ? $_[0] : $_[1] };
  my $great = sub { $_[0] > $_[1] ? $_[0] : $_[1] };
  # which block borders on this block?
  my $tx1; my $ty1; my $tx2; my $ty2;
  if($dest->x1() - 1 == &x2) { 
    # east of us
    $tx1 = $tx2 = $dest->x1();
    $ty1 = $great->(&y1, $dest->y1());
    $ty2 = $least->(&y2, $dest->y2());
  } elsif($dest->x2() + 1 == &x1) { 
    # west of us
    $tx1 = $tx2 = $dest->x2(); 
    $ty1 = $great->(&y1, $dest->y1());
    $ty2 = $least->(&y2, $dest->y2());
  } elsif($dest->y1() - 1 == &y2) {
    # south of us
    $ty1 = $ty2 = $dest->y1();
    $tx1 = $great->(&x1, $dest->x1());
    $tx2 = $least->(&x2, $dest->x2());
  } elsif($dest->y2() + 1 == &y1) {
    # north of us
    $ty1 = $ty2 = $dest->y2();
    $tx1 = $great->(&x1, $dest->x1());
    $tx2 = $least->(&x2, $dest->x2());
  } else {
    warn "*** these blocks aren't touching, whats wrong with you!? blocks: " .
         $me->get_color() . ' and ' . $dest->get_color() . 
         sprintf " our: %d,%d-%d,%d their: %d,%d-%d,%d",
         &x1, &y1, &x2, &y2, $dest->x1, $dest->y1, $dest->x2, $dest->y2;
    return;
  }
  return ($tx1, $tx2, $ty1, $ty2);
}

sub beepoint {
  my $me = $_[0];
  my $dest = $_[1];
  my $x = $_[2];
  my $y = $_[3];
  # $tx1,$ty1 -> $tx2,$ty2 describes the rectangle we're walking towards
  (my $tx1, my $tx2, my $ty1, my $ty2) = $me->common_border($dest);
  # dont go sideways at all if we're already aligned
  $tx1 = $tx2 = $x if($x >= $tx1 and $x <= $tx2);
  $ty1 = $ty2 = $y if($y >= $ty1 and $y <= $ty2);
  # we normalize this to the single point nearist us: $tx1, $ty1
  $tx1 = $tx2 if($x > $tx2);
  $ty1 = $ty2 if($y > $ty2);
  return ($tx1, $ty1);
}

sub beeline {

  my $me = $_[0];
  my $dest = $_[1];
  my $x = $_[2];
  my $y = $_[3];

  # go from x, y to whereever we need to go to to make it into block $dest
  my $tx1; my $ty1;
  if($dest) {
    ($tx1, $ty1) = $me->beepoint($dest, $x, $y);
  } else {
    # silly special case: we're almost there, and the target is middle of curr block
    $dest = $me;
    $tx1 = $_[4]; $ty1 = $_[5];
  }

  # print "debug: beepoint told beeline: $tx1, $ty1\n";

  my @output;
  my $safetest = sub { $me->bounded(@_) or $dest->bounded(@_) };
  # note that N/S are reversed...
  my $gonorth = sub { return unless $safetest->($x, $y-1); $y--; push @output, 'S' };
  my $gosouth = sub { return unless $safetest->($x, $y+1); $y++; push @output, 'N' };
  my $gowest =  sub { return unless $safetest->($x-1, $y); $x--; push @output, 'W' }; 
  my $goeast =  sub { return unless $safetest->($x+1, $y); $x++; push @output, 'E' };
  my $sx = $tx1 < $x ? $gowest : $goeast;
  my $sy = $ty1 < $y ? $gonorth : $gosouth;

  my $absdif = sub { $_[0] > $_[1] ? $_[0] - $_[1] : $_[1] - $_[0] };
  my $xdiff = sub { $absdif->($x, $tx1) };
  my $ydiff = sub { $absdif->($y, $ty1) };

  0 and print "disclosure: current: $x, $y  target: $tx1, $ty1\n",
        $tx1 < $x ? "we want to deinc current x\n" : "we want to inc current x\n",
        $ty1 < $y ? "we want to deinc current y\n" : "we want to inc current y\n";

  while($x != $tx1 or $y != $ty1) {
    0 and do {
      eval { use PrettyMap; };
      print "debug: 2: $x, $y\n";
      PrettyMap::print($Block::Builder::globmap, [$x, $y, '%']);
    };
    if($xdiff->() >= $ydiff->()) {
      $sx->() or do { $ydiff->() and $sy->(); };
    } elsif($xdiff->() < $ydiff->()) {
      $sy->() or do { $xdiff->() and $sx->(); };
    }
  }

  return @output;

}

sub bounded {
  my $i = shift;
  my $x = shift;
  my $y = shift;
  $x >= $i->[0] and $x <= $i->[1] and $y >= $i->[2] and $y <= $i->[3];
}

1;


package Block::Builder;

use warnings;

# my $debug; use Nark; Nark::nark(sub { print @_, "\n" if $debug } );

sub plot_route {
  my $me = shift;
  my $sx = shift; my $x = $sx;
  my $sy = shift; my $y = $sy;
  my $dx = shift;
  my $dy = shift;
  my $o1 = $me->bounded_by($sx, $sy) or return;
  my $o2 = $me->bounded_by($dx, $dy) or return;
  my @blocks = $o1->solve_maze($o2);
  @blocks or do { warn "plot_route: no solution"; return; };
  my @res;
print "debug: 1: ", scalar @blocks, " blocks to go\n";
  while(scalar @blocks > 1) {
print "debug: 2: ", scalar @blocks, " blocks to go\n";
    push @res, $blocks[0]->beeline($blocks[1], $x, $y);
    ($x, $y) = $blocks[0]->beepoint($blocks[1], $x, $y);
    shift @blocks;
  }
  # walk to middle of current block...
  push @res, $blocks[0]->beeline(undef, $x, $y, $dx, $dy);
  return @res;
}

sub bounded_by {
  my $me = shift;
  my $x = shift; defined $x or die;
  my $y = shift; defined $y or die;
  foreach my $i (@$me) {
    # printf "$x $y: considering %d,%d -> %d,%d\n", map { $i->[$_] } (0,2,1,3);
    return $i if $x >= $i->[0] and $x <= $i->[1] and $y >= $i->[2] and $y <= $i->[3];
  }
  return;
}

sub for_color {
  my $me = shift;
  my $color = shift;
  foreach my $i (@$me) {
    return $i if $i->get_color() eq $color;
  }
  return;
}

sub test {
  @ARGV=qw(map.txt); my @map = map { chomp; $_ } <>;
  my $bb = new Block::Builder(\@map);
  return $bb;
}

sub land { $_[0] eq '.' or $_[0] eq '@' }

sub dump_box { sprintf "%d,%d - %d,%d", map { defined $box->[$_] ? $box->[$_] : -1 } (0,2,1,3); }

sub nonintersect {

  my $left = shift or die;
  my $right = shift or die;
  my $match = shift or die;
  my $notmatch = shift or die;

  foreach my $i (keys %$left) {
    if(exists $right->{$i}) {
      $match->($left->{$i}, $right->{$i})
    } else {
      $notmatch->($left->{$i});
    }
  }
 
}

sub new {

  my $type = shift;

  $globmap = $_[0]; # bleah

  my @map = @{shift()};


  my $me = [];

  bless $me, $type;
  
  my $y = 0;
  my $edges = {};
  my @boxes;
  my $color = 1;
  my $linelen = length($map[0]);
  
  @map = map { '#'.$_ } @map;
  # unshift @map, '#' x length($map[0]);
  push @map, '#' x length($map[0]);
  
  foreach my $line (@map) {
  
    # print $line, "\n";
  
    my $curedges = {};
    my $x = 0;
  
    $y++; 
  
    while($x < $linelen) {
      # find starting edge
      my $start = $x;
      while(!land(substr $line, $x, 1) and $x < $linelen) {
        $x++;
        $start = $x;
      }
      # find trailing edge
      while(land(substr $line, $x, 1) and $x < $linelen) {
        $x++;
      }
      $curedges->{$start.','.($x-1)} = new Block($start, $x-1, $y, undef, $color++, [], $me);
    }
  
    # debug
    foreach my $i (values %$curedges) {
      # print join ' ', map { defined $i->[$_] ? $i->[$_] : '?' } 0..3; print "\n";
    }
  
    # if a new edge is found among the old edges, we update it with the first-seen start line
    # if an old edges is not found amoung the new edges, we finalize it and forget about it
    nonintersect($curedges, $edges, sub {
      # found
      my $new = shift or die;
      my $old = shift or die;
      $new->[2] = $old->[2];
      # print "found: ", dump_box($old); # debug
    }, sub { });
    nonintersect($edges, $curedges, sub { }, sub {
      # not found
      my $old = shift or die;
      $old->[3] = $y - 1;
      push @boxes, $old;
      # print "not found: ", dump_box($old); # debug
    });
  
    $edges = $curedges; 
  
  }

  if(0) {
    
    # debugging...
    print "here comes boxes:\n";
  
    foreach my $i (@boxes) {
      # print "box: $i->[0], $i->[2] -> $i->[1], $i->[3]\n";
      (my $x1, my $x2, my $y1, my $y2) = @$i;
      my $let = $i->get_color();
      print "$x1,$y1 -> $x2,$y2\n";
      map { my $y = $_;  map { substr($map[$y], $_, 1) = $let; } $x1..$x2 } $y1-1..$y2-1;
    }

    foreach my $i (@map) { print $i, "\n"; }

  }
  
  foreach my $i (@boxes) {
    my $x1 = $i->[0] - 1;
    my $x2 = $i->[1] + 1;
    my $y1 = $i->[2] - 1;
    my $y2 = $i->[3] + 1;
    #my $xrange = sub { $great->($_[0], $x1) < $least->($_[1], $x2) };
    #my $yrange = sub { $great->($_[0], $y1) < $least->($_[1], $y2) };
    my $xrange = sub { $_[1] < $x1 and return 0; $_[0] > $x2 and return 0; 1; };
    my $yrange = sub { $_[1] < $y1 and return 0; $_[0] > $y2 and return 0; 1; };
    foreach my $j (@boxes) {
      next if $i eq $j;
      if(($j->[0] == $x2 and $yrange->($j->[2], $j->[3])) or   # east of us
         ($j->[1] == $x1 and $yrange->($j->[2], $j->[3])) or   # west
         ($j->[2] == $y2 and $xrange->($j->[0], $j->[1])) or   # south
         ($j->[3] == $y1 and $xrange->($j->[0], $j->[1]))      # north
      ) {
        $i->add_neighbor($j);
        # printf "%s borders %s\n", $lets->[$i->[4] % @$lets], $lets->[$j->[4] % @$lets];
      }
    }
  }

  push @$me, @boxes;
  return $me;
  
}
  
1;
