package SortPackages;

# take a Treasure::Chest reference. 
# decide on a combination of objects to try to deliver.
# this combination will have:
# 1. closely clustered target destinations
# 2. use close to our max carry
# 3. few packages

# some optimizations:
# 1. not too few packages for fear of dropping them
# 2. look for packages all in one or two blocks
# 3. near any packages we are already carrying...

sub go {
  my $chest = shift;
  my $blockbuilder = shift;
  my $max = shift;
  my $x = shift;
  my $y = shift;

  my $curblock = $blockbuilder->bounded_by($x, $y);

  print "debug: curblock ", $curblock->color(), "\n"; # debug
  print "debug: our current x, y: $x,$y\n"; # debug

  my @stuff;      # $stuff[$blockcolor] = [$block, [pack, pack,...], $desireability]

  foreach my $i (@$chest) {
    # print "** debug: SOMETHING in chest atleast...\n";
    next unless $i->[0] == $x and $i->[1] == $y; # only local packages
    # print "** debug: and its local too!\n";
    my $block = $blockbuilder->bounded_by($i->destx, $i->desty);
    next unless $block; # no deliveries to ponds
    # print "** debug: and its not going to a pond!\n";
    my $color = $block->color();
    # print "** debug: adding to block $color\n";
    $stuff[$color] ||= [$block, [], 0];
    push @{$stuff[$color]->[1]}, $i;
  }

  foreach my $i (@stuff) {
    next unless $i;
    next unless scalar @{$i->[1]};
    my $rating = 0;
    my $block = $i->[0];
    my $area = $block->query_x_dim() * $block->query_y_dim();
    my $packs = $i->[1];
    my $num = scalar @$packs;
    my $totweight; map { $totweight += $_->weight() } @$packs;
    my $distance = $curblock->approx_cost($block);
    $rating -= $area/$num;
    $rating -= $distance * 2;
    $rating += $totweight/$num;
    $rating -= $num;
    # print "** debug: came up with rating $rating for block $color\n";
    $i->[2] = $rating;
  }

  @stuff = sort { $b->[2] <=> $a->[2] } grep { $_ } @stuff;

  # add our top picks for as many as we can carry, then grab whatever we can carry at all

  my @resolve;

  foreach my $i (@stuff) {
    next unless $i;
    foreach my $j (@{$i->[1]}) {
      next unless $j;
      # print "** debug: considering t id " . $j->id() . "\n";
      if($j->weight() < $max) {
        # print "** we can carry that. queueing.\n";
        $max -= $j->weight();
        push @resolve, $j;
      }
    }
  }

  1 and do {
    print "** decided to take treasure:\n";
    foreach my $i (@resolve) {
      print $i->id(), " ";
    }
    print "\n";
  };

  return @resolve;

}

1;

