#!/usr/bin/perl

use Socket;
use PrettyMap;
use Block;
use SortPackages;
use Utils;
use warnings;

$|=1;

#
# config:
#

scalar @ARGV == 2 or do {
  print "usage: $0 <host> <port>\n";
  exit 1;
};

my $host = shift @ARGV;
my $port = shift @ARGV;

my $nullcmd = '1 Drop 32000';

#
# globals:
#

my @map;
my $boardx;
my $boardy;

my @bases;     # [x, y, visitedf]
my @bots;      # [x, y, carry, id]
my @treasure;  # [x, y, bot, weight, destx, desty]

my $bot; # [x,y,carry] - reference to member of @bots
my $botid;
my $carry;
my $money;

my $reeval = sub { }; # defined later to pick destination and queue it

# containers:

my $treasure = new Treasure::Chest(\@treasure) or die;
my $bases = new Base::Chest(\@bases) or die;
my $blockbuilder;  # will contain Block::Builder

# setup:

sub grepchar {
  my $test = shift;
  map { my $y = $_; map { $test->(substr($map[$y],$_,1),$_+1,$y+1) } 0..length $map[0] } 0..$#map;
}

#
# debugging cruft
#

my $debug = 1;
my $lastcmd = '';
my $debugtarget = '';

sub dump_bases {
  foreach my $i (@bases) {
    print "base at ", $i->[0], ",", $i->[1], "\n";
  }
}

sub debug {

  print "\n" x 3;

  # foreach my $i (grep {ref $_} @treasure)  {
  #   printf "t: id: %d weight: %d curloc: %d,%d destination: %d,%d\n",
  #           $i->id, $i->weight, $i->x, $i->y, $i->destx, $i->desty;
  # }

  my @objs;
  foreach my $i (grep {ref $_} @bases)     { push @objs, [$i->[0], $i->[1], '@'] }
  foreach my $i (grep {ref $_} @bots)      { push @objs, [$i->[0], $i->[1], 'R'] }
  foreach my $i (grep {ref $_} @treasure)  { push @objs, [$i->[0], $i->[1], 't'] unless $i->[2] }
  PrettyMap::print(\@map, @objs);

  print "robotid: $botid  money: $money  last command: '$lastcmd'\n";
  printf "carry: %d of %d\n", $bot->[2], $carry;
  printf "location: current: %d,%d\n", $bot->[0], $bot->[1];
  print $debugtarget;

}

$SIG{qq{__DIE__}} = sub {
  print join "\n", @_, join '', map {
    (caller($_))[0] ? sprintf("%s at line %d\n", (caller($_))[1,2]) : '';
  } (1..30);
  exit 1;
};
$SIG{qq{INT}} = $SIG{qq{__DIE__}};


#
# agent utils
#

sub plot_route {
  my $sx = shift; my $sy = shift;
  my $fx = shift; my $fy = shift;
  return $blockbuilder->plot_route($sx, $sy, $fx, $fy);
}

sub approx_cost {
  # quick metric how far/painful it is to go somewhere
  my $sx = shift; my $sy = shift;
  my $fx = shift; my $fy = shift;
  my $o1 = $blockbuilder->bounded_by($sx, $sy) or return;
  my $o2 = $blockbuilder->bounded_by($fx, $fy) or return;
  return $o1->approx_cost($o2);
}

sub detect_tunnel {
  my $sx = shift; my $sy = shift;
  my $fx = shift; my $fy = shift;
  my $o1 = $blockbuilder->bounded_by($sx, $sy) or return;
  my $o2 = $blockbuilder->bounded_by($fx, $fy) or return;
  my @route = $o1->solve_maze($o2) or return;
  my $currroom = shift @route or return;
  my $nextroom = shift @route or return;
  my $lastroom = shift @route or return;
  my @neighbors = $o1->query_neighbors();
  return if(@neighbors != 2); # more than two neighbors, not a tunnel
  return if($nextroom->query_x_dim() != 1 and $nextroom->query_y_dim() != 1);
  my $crowd = 0;
  foreach my $i (@bots) {
    $crowd++ if $nextroom->bounded($i->[0], $i->[1]) or $lastroom->bounded($i->[0], $i->[1]);
  }
  return $crowd;
}

sub server_response {
  # #1 X 1 Y 1 E #2 E 
  my $arg = shift;
  my @tokens = split / /, $arg;
  my $curbot;
  while(my $token = shift @tokens) {
    if($token =~ m/^#([0-9]+)$/) {
      # possible new Bot
      $curbot = $1;
      $bots[$curbot]->[3] = $curbot;
      # print "debug: curbot $1\n";
    } elsif($token eq 'X') {
      # print "debug: setx $tokens[0]\n";
      $bots[$curbot]->[0] = shift @tokens;
    } elsif($token eq 'Y') {
      # print "debug: sety $tokens[0]\n";
      $bots[$curbot]->[1] = shift @tokens;
    } elsif($token eq 'N') {
      $bots[$curbot]->[1]++;
    } elsif($token eq 'S') {
      $bots[$curbot]->[1]--;
    } elsif($token eq 'W') {
      $bots[$curbot]->[0]--;
    } elsif($token eq 'E') {
      $bots[$curbot]->[0]++;
    } elsif($token eq 'P') {
      my $t = shift @tokens;
      $treasure->get($t)->set_bot($bots[$curbot]);
      $bots[$curbot]->[2] += $treasure->get($t)->weight();
      # print "debug: take\n";
    } elsif($token eq 'D') {
      my $t = shift @tokens;
      $treasure->get($t)->set_x_y($bots[$curbot]->[0], $bots[$curbot]->[1]);
      $bots[$curbot]->[2] -= $treasure->get($t)->weight();
      # print "debug: drop\n";
    } else {
      warn " unknown token: '$token'";
    }
  }
}

sub hangup {
  debug();
  print "Connection closed. Existance deallocated.\n";
  exit 0;
}

#
# network process
#

sub talk_to_server {

  # connect to server.

  my $sh;

  socket($sh, PF_INET, SOCK_STREAM, getprotobyname('tcp')) or die $!;
  connect($sh, sockaddr_in($port, inet_aton($host))) or die $!;
  select((select($sh), $|=1)[0]) or die $!; # unbuffered

  # stupid eols
  my $readsh = sub { 
    my $tmp = <$sh>; 
    chomp $tmp; 
    print "--> $tmp\n"; # debug
    if($tmp =~ m/Robot (.*) died/) { hangup(); }
    return $tmp; 
  };

  # setup

  # send string "Player\n"
  # recv board: x y\nboard....\n
  # recv config: myrobotid maxcarry money\n
  # recv player locations/serv response: robotid X n Y n ...\n, eg:
  # #2 X 2 Y 1 #1 X 1 Y 2 #3 X 1 Y 1

  print $sh "Player\n" or die $!;
  ($boardx, $boardy) = split / /, $readsh->();
  foreach my $i (0..$boardy-1) {
    $map[$i] = $readsh->();
  }

  # okey, strong evidence suggests we dont want to flip the map ;)
  # read map upsidedown? could we fix all the coordinate stupidity right here? hmm
  # @map = reverse @map;
  # or would i just have to reverse the robots etc too?

  ($botid, $carry, $money) = split / /, $readsh->();
  $bot = [undef, undef, 0, $botid]; # new Bot
  $bots[$botid] = $bot;
  server_response($readsh->());

  grepchar(sub { $bases->create($_[1], $_[2]) if $_[0] eq '@'; } );
  dump_bases() if $debug;

  #  1. Server: a list of packages available at the player's current location - or new line
  #             id destx desty weight ..., eg: 17 133 28 50 89 11 57 78
  #  2. Client: issue a command to the robot
  #             <bid> <Cmd> <args>
  #             1 Pick <id>
  #             1 Move S
  #  3. Server: server responds with what happened during the turn 
  #             N              The robot moved north.
  #             S              The robot moved south.
  #             E              The robot moved east.
  #             W              The robot moved west.
  #             P id           The robot picked up package with id id.
  #             D id           The robot dropped package with id id.
  #             X num1 Y num2  The robots location is set to (num1,num2). 
  # for example: 
  # #1 X 1 Y 1 E #2 E 
  # Robot 2 moved east and robot 1 appeared at location (1,1) and moved east.
  
  return new MessageMethod sub {

    (my $command, my $args) = @_;

    if($command eq 'packages') {
      return $readsh->();
    }

    if($command eq 'play') {
      $lastcmd = $args;
      print "**> $args\n";
      print $sh ($args . "\n");
      return;
    }

    if($command eq 'result') {
      return $readsh->();
    }

    warn "unknown command '$command'";

  };

}

#
# servlets
#

my @servlets;

# track position of other bots and treasure

push @servlets, do {
  # run first, since descion-making servlets may require our info
  new MessageMethod sub {
    my $action = shift;

    if($action eq 'notify_packages') {
      # id destx desty weight ..., eg: 17 133 28 50 89 11 57 78
      my @tmp = split / /, shift();
      while(@tmp) {
        # printf "*** debug: creating treasure record: %d %d,%d %d\n", @tmp;
        my $t = $treasure->get(shift @tmp);
        $t->x = $bot->[0];
        $t->y = $bot->[1];
        $t->destx = shift @tmp;
        $t->desty = shift @tmp;
        $t->weight = shift @tmp;
      }
      # debug
      foreach my $i (@treasure) { 
        0 and printf "*** debug: vrfy treasure record: %d %d,%d %d\n", 
               $i->id, $i->destx, $i->desty, $i->weight;
      }
      return;
    }

    if($action eq 'suggest_move') {
      return [0, sub { $nullcmd }];
    }

    if($action eq 'notify_result') {
      server_response(shift());
      return;
    }

    warn "unknown action '$action'";

  };
};

# head towards treasure at dropped location or base, or else deliver treasure
# depending on carry capacity or distance or something

push @servlets, do {

  my @presroute; my $prestreasure;

  my $attainables = sub {

     my $cap = $carry - $bot->[2];
     my @routes = ();

     # print "** debug: cap is $cap\n";

     foreach my $i ($treasure->by_bot($bot)) {
       # print "XXX debug: delivery?...\n";
       next unless $i;
       # print "XXX debug: considering delivering treasure: ", $i->id, "\n";
       # debug
       # printf "deliver?: id: %d weight: %d curloc: %d,%d destination: %d,%d\n",
       #         $i->id, $i->weight, $i->x, $i->y, $i->destx, $i->desty;
       # next unless $i->[0] and $i->[1];
       print "  .. exists\n";
       my $dropoff = $i->get_dropoff();
       my @coords = ($bot->[0], $bot->[1], $dropoff->x(), $dropoff->y());
       my $route = [$dropoff, sub { plot_route(@coords) }, undef ];
       $route->[2] = approx_cost(@coords);
       $route->[2] -= 100 if($cap < $carry / 2);
       $route->[2] -= 100 if($cap < $carry / 3);
       $route->[2] -= 100 if($cap < $carry / 4);
       push @routes, $route;
     };

     foreach my $i (@$treasure) {
       next unless $i;
       next unless $i->[0] and $i->[1];
       next unless $i->weight() <= $cap;
       next if $i->bot();
       # stop trying to plot a course to the present spot!
       next if $i->x == $bot->[0] and $i->y == $bot->[1]; 
       my @coords = ($bot->[0], $bot->[1], $i->x(), $i->y());
       my $route = [$i, sub { plot_route(@coords) }, undef];
       $route->[2] = approx_cost(@coords);
       $route->[2] += 20 if($cap < $carry / 2);
       push @routes, $route;
     }

     foreach my $i (@bases) {
       my @coords = ($bot->[0], $bot->[1], $i->[0], $i->[1]);
       my $route = [$i, sub { plot_route(@coords) }, undef];
       $route->[2] = approx_cost(@coords);
       # dont go the base itself, go to the treasure! this kicks in after weve been there
       if($i->visited()) {
         $route->[2] += 1000;
       } else {
         $route->[2] -= 100; 
       }
       $route->[2] += 1000 if($cap < $carry /2);
       push @routes, $route;
     }

     # XX should sort them according to distance/value/chance of getting metrix, not just dist

     @routes = sort { $a->[2] <=> $b->[2] } @routes;

     return @routes;

  };

  my $possiblebasevisit = sub {
     foreach my $i (@bases) {
       next unless $i and $i->isa('Base');;
       $i->visited = 1 if $i->x == $bot->[0] and $i->y == $bot->[1];
     }
  };

  my $picktarget = sub {
    $prestreasure = undef;
    @presroute = ();
    my @routes = $attainables->();
    # debug
    foreach my $i (@routes) {
      printf "** debug: route: weight: %d type: %s\n",
             $i->[2], ref $i->[0];
    }
    if(scalar @routes) {
      $prestreasure = $routes[0]->[0];
      @presroute = $routes[0]->[1]->();
    } else {
      print "*** debug: no attainable routes - at all!\n";
    }
    $debugtarget = sprintf "target: type: %s location: %d, %d\nroute to targ: '%s'\n", 
      ref $prestreasure, $prestreasure->[0], $prestreasure->[1],
      join ' ', @presroute;
  };

  $reeval = $picktarget;

  new MessageMethod sub {
    my $action = shift;
    if($action eq 'notify_packages') {
      return;
    }
    if($action eq 'suggest_move') {
      if(@presroute and scalar @presroute) {
        return [50, sub { my $dir = shift @presroute; return "1 Move $dir"; }];
      } else {
        return [2, sub { $reeval->(); $nullcmd }];
      }
    }
    if($action eq 'notify_result') {
      $possiblebasevisit->();
      if($prestreasure and $prestreasure->bot()) {
        # someone stole it!
        $reeval->();
      } elsif(!@presroute) {
        # reached the end of the route
        $reeval->();
      }
      return;
    }
    warn "unknown action '$action'";
  };
};

# pick up packages

push @servlets, do {
  my @goodies;
  my $goodiesx = 0; my $goodiesy = 0;
  new MessageMethod sub {
    my $action = shift;
    if($action eq 'notify_packages') {
      my $arg = shift;
      return unless $arg;
      $goodiesx = $bot->[0];
      $goodiesy = $bot->[1];
      @goodies = SortPackages::go(
        $treasure, $blockbuilder, $carry-$bot->[2], $bot->[0], $bot->[1]
      );
      return;
    }
    if($action eq 'suggest_move') {
      if(@goodies) {
        return [80, sub { 
          my $p = shift(@goodies); 
          $reeval->() unless @goodies;
          '1 Pick ' . $p->id(); 
        } ];
      } else {
        return [0, sub { $nullcmd }];
      }
    }
    if($action eq 'notify_result') {
      return;
    }
    warn "unknown action '$action'";
  };
};

# at a delivery location? drop the package!

push @servlets, do {
  new MessageMethod sub {
    my $action = shift;
    if($action eq 'notify_packages') {
      return;
    }
    if($action eq 'suggest_move') {
      foreach my $i ($treasure->by_bot($bot)) {
        if($i->destx == $bot->[0] and $i->desty == $bot->[1]) {
          my $id = $i->id();
          return [90, sub { 
            $treasure->zero($id);
            "1 Drop $id" 
          }];
        }
      }
      return [0, sub { $nullcmd }];
    }
    if($action eq 'notify_result') {
      return;
    }
    warn "unknown action '$action'";
  };
};

# move packages off base where bots are going to be slightly less likely to notice them,
# preferably into one big cache thats in the opposite direction of any base

push @servlets, do {

  my @route = ();
  my $targetx = 0;
  my $targety = 0;

  my $instagate = sub {
    my $o1 = $blockbuilder->bounded_by($bot->[0], $bot->[1]);
    $targetx = $o1->x2;
    $targety = $o1->y2;
    @route = plot_route($bot->[0], $bot->[1], $targetx, $targety);
  };

  my $dropstuff = sub {
    foreach my $i ($treasure->by_bot($bot)) {
      return "1 Drop " . $i->id();
    }
    $reeval->();
    return $nullcmd;
  };

  new MessageMethod sub {
    my $action = shift;
    if($action eq 'notify_packages') {
      return;
    }
    if($action eq 'suggest_move') {
      if(@route) {
        return [70, sub { my $move = shift @route; $reeval->() unless @route; "1 Move $move" }];
       } elsif($bot->[0] == $targetx and $bot->[1] == $targety) {
        return [95, $dropstuff ];
      } else {
        return [0, sub { $nullcmd }];
      }
    }
    if($action eq 'notify_result') {
      # are the conditions right?
      my $numbots; foreach my $i (@bots) { $numbots++ if($i) }
      my $t; foreach my $i (@$treasure) { $t++ }
      $instagate->() if($numbots > 1 and $t/$numbots > 2); 
      return;
    }
    warn "unknown action '$action'";
  };
};

# delete bogus treasure
# okey that was a bad idea

#  push @servlets, do {
#    new MessageMethod sub {
#      my $action = shift;
#      if($action eq 'notify_packages') {
#        my $packs = shift;
#        return if $packs;
#        foreach my $i (@$treasure) {
#          # this was dropped, possibly by us, and it vanished. get rid of it. let it trouble us
#          # no more.
#          next unless $i;
#          next if $i->bot();
#          if($i->x == $bot->[0] and $i->y == $bot->[1]) {
#            $i = undef;
#          }
#        }
#        return;
#      }
#      if($action eq 'suggest_move') {
#        return [0, sub { $nullcmd }];
#      }
#      if($action eq 'notify_result') {
#        return;
#      }
#      warn "unknown action '$action'";
#    };
#  };

# push other bots and steal their treasure, even if we have to burn a few turns

# stay out of really bum-fuck situations... too many bots in a tunnel.
# instead, sit at the end of the tunnel and steal treasure from bots as they come out.

# is another bot near? try to time it so that we push them

# place packages by water, wait for bots to come and get it then push them, bwahahaha!

# empty agent, ripe for the cutting and pasting

push @servlets, do {
  new MessageMethod sub {
    my $action = shift;
    if($action eq 'notify_packages') {
      return;
    }
    if($action eq 'suggest_move') {
      return [0, sub { $nullcmd }];
    }
    if($action eq 'notify_result') {
      return;
    }
    warn "unknown action '$action'";
  };
};

# give graphical output as a debugging aid for poor scott as we test this puppy 

push @servlets, do {
  # dump pretty map
  new MessageMethod sub {
    my $action = shift;
    if($action eq 'notify_packages') { return; }
    if($action eq 'suggest_move') { return [0, sub { $nullcmd } ] }
    if($action eq 'notify_result') { debug(); return; }
    warn "unknown action '$action'";
  };
};

#
# main routine, i guess, for lack of threads:
#

my $con = talk_to_server();

$blockbuilder = new Block::Builder(\@map);

$reeval->(); # plot routes to bases right off 

while(1) {

  # ->notify_packages()
  # ->suggest_move()
  # ->notify_result(0

  my $packs = $con->packages();
  map { $_->notify_packages($packs) } @servlets;

  my @moves = sort { $b->[0] <=> $a->[0] } map { $_->suggest_move() } @servlets;
  print "debug: command priority: ", $moves[0]->[0], "\n";
  $con->play($moves[0]->[1]->());

sleep 1; # debug XXXX

  my $result = $con->result();
  hangup() if $result eq 'exit';
  map { $_->notify_result($result) } @servlets;
  
}

